module StateMachine
  module Integrations #:nodoc:
    module ActiveModel
      # Adds support for invoking callbacks on ActiveModel observers with more
      # than one argument (e.g. the record *and* the state transition).  By
      # default, ActiveModel only supports passing the record into the
      # callbacks.
      # 
      # For example:
      # 
      #   class VehicleObserver < ActiveModel::Observer
      #     # The default behavior: only pass in the record
      #     def after_save(vehicle)
      #     end
      #     
      #     # Custom behavior: allow the transition to be passed in as well
      #     def after_transition(vehicle, transition)
      #       Audit.log(vehicle, transition)
      #     end
      #   end
      module Observer
        def self.included(base) #:nodoc:
          base.class_eval do
            alias_method :update_without_multiple_args, :update
            alias_method :update, :update_with_multiple_args
          end
        end
        
        # Allows additional arguments other than the object to be passed to the
        # observed methods
        def update_with_multiple_args(observed_method, object, *args) #:nodoc:
          if args.any?
            send(observed_method, object, *args) if respond_to?(observed_method)
          else
            update_without_multiple_args(observed_method, object)
          end
        end
      end
    end
  end
end

ActiveModel::Observer.class_eval do
  include StateMachine::Integrations::ActiveModel::Observer
end if defined?(ActiveModel::Observer)
